const _ = require('lodash')
const chokidar = require('chokidar')
const del = require('del')
const fs = require('fs')
const globby = require('globby')
const makeDir = require('make-dir')
const importer = require('./transforms/importer')
const matter = require('gray-matter')
const MarkdownIt = require('markdown-it')
const path = require('path')

/**
 * Container for MarkdownIt
 */
let md

/**
 * Config with defaults
 * @type {Object}
 */
const CONFIG = {
  src: 'content',
  out: 'api',
  debug: false,
  markdown: {
    html: true
  }
}

/**
 * Data store
 * @type {Object}
 */
const STORE = {
  paths: [],
  updated: [],
  nodes: {},
  files: []
}

/**
 * Transforms middleware
 * @type {Array}
 */
const TRANSFORMS = [ importer ]

/**
 * Build api from content files
 * @param  {Object} config
 * @return {Promise}
 */
async function build (config) {
  _.merge(CONFIG, config)

  configureMarkdown()

  try {
    STORE.paths = await getFiles()
    STORE.nodes = await buildNodes()

    await applyMiddleware()
    await prepareOutputDir()
    await writeNodes()
    await writeRoutesFile()
    await writeFiles()

  } catch (err) {
    console.error(err)
  }
}

/**
 * Configure markdown
 */
function configureMarkdown () {
  let config = CONFIG.markdown
  let plugins = config.plugins
  delete config.plugins

  md = new MarkdownIt(config)

  if (_.isFunction(plugins)) plugins(md)

  if (_.isArray(plugins)) {
    plugins.forEach(plugin => {
      md.use(plugin)
    })
  }
}

async function watch (config) {
  await build(config)

  const watcher = chokidar.watch([ path.join(CONFIG.src, '**/*') ])

  console.log(`Listening for changes to files in /${CONFIG.src}/`)

  watcher.on('change', async files => {
    try {
      await update([ files ])
    } catch (err) {
      console.error(err)
      process.exit(1)
    }
  })
}

/**
 * Update nodes
 *
 * TODO:
 * - When files are REMOVED, make sure they're deleted from the output directory
 * - When files are ADDED, make sure they're added...
 *
 * @param  {Array} filepaths Paths to files that have changed
 * @return {Promise}
 */
async function update (filepaths) {
  STORE.updated = filepaths

  try {
    STORE.nodes = await buildNodes(STORE.updated)

    await applyMiddleware()

    // TODO
    // - Currently files are being written for all nodes on each update
    // - Only write files for nodes that have changed
    const results = await writeNodes(filepaths)

    await writeRoutesFile()

    // TODO: make this output prettier
    console.log(`Updated: \n${results}`)
    STORE.updated = []
  } catch (err) {
    console.error(err)
    process.exit(1)
  }
}

/**
 * Get files
 * @return {Promise}
 */
async function getFiles () {
  try {
    return await globby(path.join(CONFIG.src, '**/*'))
  } catch (err) {
    console.error(err)
  }
}

/**
 * Build nodes from STORE.paths
 * @param  {Array} files
 * @return {Promise}
 */
async function buildNodes (paths = STORE.paths) {
  const q = []
  const nodeMap = {}

  paths.forEach(filePath => {
    q.push(buildNode(filePath))
  })

  const nodes = await Promise.all(q)

  nodes.forEach(node => {
    if (!node) return
    nodeMap[node.path] = node
  })

  return nodeMap
}

/**
 * Build an individual node
 * @param  {String} filePath
 * @return {Promise}
 */
async function buildNode (filePath) {
  const contentPath = filePath.replace(CONFIG.src, '')
  let contents = await readFile(filePath)

  // Handle yaml files
  if (_.endsWith(filePath, '.yaml')) contents = `---\n${contents}`

  // Handle pass-through files
  if (!_.includes(['.md', '.yaml'], path.extname(filePath))) {
    STORE.files.push({
      src: filePath,
      out: path.join(CONFIG.out, contentPath),
      contents
    })

    return
  }

  const parsed = matter(contents)

  return {
    path: contentPath,
    attributes: parsed.data,
    body: md.render(parsed.content)
  }
}

/**
 * Apply middleware
 * @return {Promise}
 */
async function applyMiddleware () {
  const { config, store } = await applyTransforms()
  _.merge(CONFIG, config)
  _.merge(STORE, store)
}

/**
 * Apply transforms
 * @param  {Array} nodes
 * @return {Promise}
 */
async function applyTransforms () {
  for (const task of TRANSFORMS) {
    q = await task({ config: CONFIG, store: STORE })
  }

  return q
}

/**
 * Prepare the output directory
 * @return {Promise}
 */
async function prepareOutputDir () {
  const patterns = [
    path.join(CONFIG.out, '**'),
    `!${CONFIG.out}`
  ]

  // Ensure the directory exists
  await makeDir(CONFIG.out)

  return del(patterns)
}

/**
 * Write nodes to the filesystem
 * @param  {Array} filepaths If passed, only write nodes for these
 * @return {Promise}
 */
function writeNodes (filepaths) {
  let nodes = STORE.nodes
  const q = []

  if (filepaths) {
    nodes = {}

    // Limit to updated nodes
    filepaths.forEach(filepath => {
      let key = filepath.replace(CONFIG.src, '')
      nodes[key] = STORE.nodes[key]
    })
  }

  _.forEach(nodes, (node, nodePath) => {
    q.push(writeNode(node, nodePath))
  })

  return Promise.all(q)
}

/**
 * Write file to output directory
 * @param  {Object} node
 * @param  {String} nodePath
 * @return {Promise}
 *
 * TODO: clean this up
 */
async function writeNode (node, nodePath) {
  const basename = path.basename(nodePath)
  const extname = path.extname(nodePath)
  const filename = path.basename(nodePath, extname)
  let writePath
  let out

  writePath = path.join(CONFIG.out, nodePath.replace(basename, ''))

  node.path = node.path.replace(basename, '')

  try {
    await makeDir(writePath)

    out = _.cloneDeep(node)

    // Remove any attribute starting with an underscore
    _.forEach(out.attributes, (attribute, key) => {
      if (_.startsWith(key, '_')) delete out.attributes[key]
    })

    if (basename === 'index.md') {
      writePath = path.join(writePath, 'index.json')
    } else {
      writePath = path.join(writePath, `${filename}.json`)
      out = out.attributes
    }

    return await writeFile(writePath, JSON.stringify(out, null, CONFIG.debug ? 4 : 0))
  } catch (err) {
    console.error(err)
  }
}

/**
 * Write routes file
 * @return {Promise}
 */
async function writeRoutesFile () {
  const writePath = path.join(CONFIG.out, 'routes.json')
  const out = { routes: _.uniq(_.map(STORE.nodes, 'path')) }

  await writeFile(writePath, JSON.stringify(out, null, CONFIG.debug ? 4 : 0))
}

/**
 * Write pass-through files
 * @return {Promise}
 */
function writeFiles () {
  const q = []

  STORE.files.forEach(file => {
    q.push(writeFile(file.out, file.contents))
  })

  return Promise.all(q)
}

/**
 * Promisify fs.readFile
 * @param {String} filePath
 * @return {Promise}
 */
function readFile (filePath) {
  return new Promise((resolve, reject) => {
    fs.readFile(filePath, 'utf8', (err, data) => {
      if (err) reject(err)
      resolve(data)
    })
  })
}

/**
 * Promisify fs.writeFile
 * @param {String} filePath
 * @param {Object} content
 * @return {Promise}
 */
function writeFile (filePath, content) {
  return new Promise((resolve, reject) => {
    fs.writeFile(filePath, content, 'utf8', err => {
      if (err) reject(err)
      resolve(filePath.replace(CONFIG.out, ''))
    })
  })
}

module.exports.config = CONFIG
module.exports.store = STORE
module.exports.build = build
module.exports.watch = watch
