const _ = require('lodash')
const micromatch = require('micromatch')
const path = require('path')

const fs = require('fs')

/**
 * Data store (to be merged with main data store)
 * @type {Object}
 */
const STORE = {
  dependencies: {},
  nodes: {}
}

const CONFIG = {}

/**
 * Initializer
 * @param  {object} context
 * @return {Array}
 */
module.exports = async function (context) {
  _.merge(CONFIG, context.config)
  _.merge(STORE, context.store)

  let nodes = STORE.nodes

  // If there are items in the update queue, only act on those items
  if (STORE.updated.length) {
    nodes = {}

    _.forEach(STORE.nodes, (node, key) => {
      if (_.includes(STORE.updated, path.join(CONFIG.src, key))) {
        nodes[key] = node
      }
    })
  }

  await mapImports(nodes)

  if (STORE.updated.length) {
    await mapDependencyImports()
  }

  _.merge(context.store, STORE)

  return context
}

async function mapDependencyImports () {
  const updated = []
  const nodes = {}

  _.forEach(STORE.dependencies, (dependencies, key) => {
    _.forEach(STORE.updated, item => {
      // If item depends on something from STORE.updated, re-map imports for this item
      if (!_.includes(dependencies, item.replace(CONFIG.src, ''))) return
      updated.push(key)
      nodes[key] = STORE.nodes[key]
    })
  })

  // Add updated items to STORE.updated, so that they are written out to the filesystem
  STORE.updated = _.union(STORE.updated, updated)

  // Re-map imports for these updated nodes
  await mapImports(nodes)
}

/**
 * Map imports
 * @param {Object} nodes All nodes or only updated nodes
 * @return {Promise}
 */
async function mapImports (nodes) {
  const nodeMap = {}
  const q = []

  _.forEach(nodes, node => {
    // Create dependency array for this node
    STORE.dependencies[node.path] = []

    // Exit here if no imports
    if (!node.attributes || !node.attributes._import) {
      nodeMap[node.path] = node
      return
    }

    _.forEach(node.attributes._import, (importObj, importKey) => {
      q.push(applyImport(node, importObj, importKey))
    })
  })

  nodes = await Promise.all(q)

  // Convert array returned from promise into object
  nodes.forEach(node => {
    nodeMap[node.path] = node
  })

  // Remove empty dependencies
  _.forEach(STORE.dependencies, (dep, key) => {
    if (!dep.length) delete STORE.dependencies[key]
  })

  // Merge nodeMap into store
  _.assign(STORE.nodes, nodeMap)
}

/**
 * Apply an individual import for a node
 * @param  {Object} node
 * @param  {Object} importObj The import definition
 * @param  {String} importKey Name of the import array
 * @return {Promise}
 */
async function applyImport (node, importObj, importKey) {
  const matched = micromatch(_.keys(STORE.nodes), importObj.match)

  // Exit here if no matches
  if (!matched) return node

  // Initialize imported collection
  node.attributes[importKey] = []

  matched.forEach(filename => {
    const match = STORE.nodes[filename]

    const result = {
      path: match.path.replace('index.md', ''),
      attributes: {}
    }

    // Only import whitelisted attributes
    _.forEach(match.attributes, (val, key) => {
      if (!_.includes(importObj.attributes, key)) return
      result.attributes[key] = val
    })

    // Include body if `body` is true
    if (importObj.body) {
      result.body = match.body
    }

    node.attributes[importKey].push(result)

    // Add items to dependencies graph
    STORE.dependencies[node.path].push(match.path)
  })

  node.attributes[importKey] = applyFilters(node.attributes[importKey], importObj)

  return node
}

/**
 * Apply filters to imported items
 * @param  {Array} items
 * @param  {Object} settings
 * @return {Promise}
 */
function applyFilters (items, { filter, order, sort, limit }) {
  if (filter) items = _.filter(items, { attributes: filter })
  if (order) items = _.orderBy(items, `attributes.${order}`, sort)
  if (limit) items = _.slice(items, 0, limit)

  return items
}
