const { build, config, store } = require('../src')
const fileExists = require('file-exists')
const fs = require('fs')

test('merges user config with defaults', async () => {
  await build({ src: 'foo', out: 'test/api' })

  expect(config.src).toBe('foo')
  expect(config.out).toBe('test/api')
})

test('finds source files', async () => {
  await build({ src: 'test/content', out: 'test/api' })

  expect(store.paths).toContain('test/content/index.md')
  expect(store.paths).toContain('test/content/about/index.md')
  expect(store.paths).toContain('test/content/news/index.md')
  expect(store.paths).toContain('test/content/data.yaml')
})

test('creates content nodes with required keys and values', async () => {
  await build({ src: 'test/content', out: 'test/api' })

  expect(store.nodes).toEqual(
    expect.objectContaining({
      '/index.md': expect.objectContaining({
        path: '/',
        attributes: expect.objectContaining({
          title: 'Home'
        })
      }),
      '/about/index.md': expect.objectContaining({
        path: '/about/',
        attributes: expect.objectContaining({
          title: 'About'
        })
      }),
      '/news/2017/11/17/index.md': expect.objectContaining({
        path: '/news/2017/11/17/',
        attributes: expect.objectContaining({
          title: 'Another post'
        })
      }),
      '/news/2018/04/22/index.md': expect.objectContaining({
        path: '/news/2018/04/22/',
        attributes: expect.objectContaining({
          title: 'Some blog post'
        })
      }),
      '/data.yaml': expect.objectContaining({
        path: '/',
        attributes: expect.objectContaining({
          title: 'My Site'
        })
      })
    })
  )
})

test('handles content imports', async () => {
  await build({ src: 'test/content', out: 'test/api' })

  expect(store.nodes).toEqual(
    expect.objectContaining({
      '/index.md': expect.objectContaining({
        attributes: expect.objectContaining({
          _import: expect.objectContaining({
            posts: expect.objectContaining({
              match: expect.arrayContaining([
                '/news/**/index.md'
              ])
            })
          }),
          posts: expect.arrayContaining([
            expect.objectContaining({
              path: '/news/2017/11/17/',
              attributes: expect.objectContaining({
                title: 'Another post'
              })
            }),
            expect.objectContaining({
              path: '/news/2018/04/22/',
              attributes: expect.objectContaining({
                title: 'Some blog post'
              })
            })
          ])
        })
      })
    })
  )
})

test('writes out files', async () => {
  await build({ src: 'test/content', out: 'test/api' })
  const exists = await fileExists('./test/api/index.json')

  expect(exists).toBeTruthy()
})

test('writes routes file', async () => {
  await build({ src: 'test/content', out: 'test/api' })

  expect(JSON.parse(fs.readFileSync('./test/api/routes.json'))).toEqual(
    expect.objectContaining({
      'routes': expect.arrayContaining([
        '/',
        '/about/'
      ])
    })
  )
})

test('removes attribute keys that start with an underscore', async () => {
  await build({ src: 'test/content', out: 'test/api' })
  const file = JSON.parse(fs.readFileSync('./test/api/index.json', 'utf-8'))

  expect(file.attributes._import).toBeFalsy()
})
